# Copyright (c) nexB Inc. and others. All rights reserved.
# http://nexb.com and https://github.com/nexB/scancode-toolkit/
# The ScanCode software is licensed under the Apache License version 2.0.
# Data generated with ScanCode require an acknowledgment.
# ScanCode is a trademark of nexB Inc.
#
# You may not use this software except in compliance with the License.
# You may obtain a copy of the License at: http://apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#
# When you publish or redistribute any data created with ScanCode or any ScanCode
# derivative work, you must accompany this data with the following acknowledgment:
#
#  Generated with ScanCode and provided on an "AS IS" BASIS, WITHOUT WARRANTIES
#  OR CONDITIONS OF ANY KIND, either express or implied. No content created from
#  ScanCode should be considered or used as legal advice. Consult an Attorney
#  for any legal advice.
#  ScanCode is a free software code scanning tool from nexB Inc. and others.
#  Visit https://github.com/nexB/scancode-toolkit/ for support and download.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import io
import logging
import re

import attr
from packageurl import PackageURL


TRACE = False

logger = logging.getLogger(__name__)

if TRACE:
    import sys
    logging.basicConfig(stream=sys.stdout)
    logger.setLevel(logging.DEBUG)


@attr.s()
class GoModule(object):
    namespace = attr.ib(default=None)
    name = attr.ib(default=None)
    version = attr.ib(default=None)
    module = attr.ib(default=None)
    require = attr.ib(default=None)
    exclude = attr.ib(default=None)

    def purl(self, include_version=True):
        version = None
        if include_version:
            version = self.version
        return PackageURL(
                    type='golang',
                    namespace=self.namespace,
                    name=self.name,
                    version=version
                ).to_string()


#######################################################################################
"""
This part handles go.mod files from Go.
See https://golang.org/ref/mod#go.mod-files for details

For example:

module example.com/my/thing

go 1.12

require example.com/other/thing v1.0.2
require example.com/new/thing v2.3.4
exclude example.com/old/thing v1.2.3
require (
    example.com/new/thing v2.3.4
    example.com/old/thing v1.2.3
)
require (
    example.com/new/thing v2.3.4
    example.com/old/thing v1.2.3
)

"""

"""
module is in the form
module github.com/alecthomas/participle

For example:
>>> p = parse_module('module github.com/alecthomas/participle')
>>> assert p.group('module') == ('github.com/alecthomas/participle')

require or exclude can be in the form
require github.com/davecgh/go-spew v1.1.1
or
exclude github.com/davecgh/go-spew v1.1.1
or
github.com/davecgh/go-spew v1.1.1

For example:
>>> p = parse_dep_link('require github.com/davecgh/go-spew v1.1.1')
>>> assert p.group('namespace') == ('github.com/davecgh')
>>> assert p.group('name') == ('go-spew')
>>> assert p.group('version') == ('v1.1.1')

>>> p = parse_dep_link('exclude github.com/davecgh/go-spew v1.1.1')
>>> assert p.group('namespace') == ('github.com/davecgh')
>>> assert p.group('name') == ('go-spew')
>>> assert p.group('version') == ('v1.1.1')

>>> p = parse_dep_link('github.com/davecgh/go-spew v1.1.1')
>>> assert p.group('namespace') == ('github.com/davecgh')
>>> assert p.group('name') == ('go-spew')
>>> assert p.group('version') == ('v1.1.1')
"""

# Regex expressions to parse different types of go.mod file dependency
parse_module_name = re.compile(
    r'^module\s*'
    r'(?P<namespace>[^\s]*)'
    r'\/'
    r'(?P<name>[^\s]*)\s*'
).match

parse_dep_link = re.compile(
    r'.*?(\s)*'
    r'(?P<namespace>[^\s]*)'
    r'\/'
    r'(?P<name>[^\s]*)'
    r'\s*'
    r'(?P<version>[^\s]*)\s*'
).match


def preprocess(line):
    """
    Return line string after removing commented portion and excess spaces.
    """
    if "//" in line:
        line = line[:line.index('//')]
    line = line.strip()
    return line


def parse_gomod(location):
    """
    Return a dictionary containing all the important go.mod file data.
    """
    with io.open(location, encoding='utf-8', closefd=True) as data:
        lines = data.readlines()

    gomods = GoModule()
    require = []
    exclude = []

    for i, line in enumerate(lines):
        line = preprocess(line)

        parsed_module_name = parse_module_name(line)
        if parsed_module_name:
            gomods.name = parsed_module_name.group('name')
            gomods.namespace = parsed_module_name.group('namespace')

        if 'require' in line and '(' in line:
            for req in lines[i+1:]:
                req = preprocess(req)
                if ')' in req:
                    break
                parsed_dep_link = parse_dep_link(req)
                if parsed_dep_link:
                    require.append(GoModule(
                            namespace=parsed_dep_link.group('namespace'),
                            name=parsed_dep_link.group('name'),
                            version=parsed_dep_link.group('version')
                        )
                    )
            continue

        if 'exclude' in line and '(' in line:
            for exc in lines[i+1:]:
                exc = preprocess(exc)
                if ')' in exc:
                    break
                parsed_dep_link = parse_dep_link(exc)
                if parsed_dep_link:
                    exclude.append(GoModule(
                            namespace=parsed_dep_link.group('namespace'),
                            name=parsed_dep_link.group('name'),
                            version=parsed_dep_link.group('version')
                        )
                    )
            continue

        parsed_dep = parse_dep_link(line)
        if 'require' in line:
            require.append(GoModule(
                    namespace=parsed_dep.group('namespace'),
                    name=parsed_dep.group('name'),
                    version=parsed_dep.group('version')
                )
            )
            continue

        if 'exclude' in line:
            exclude.append(GoModule(
                    namespace=parsed_dep.group('namespace'),
                    name=parsed_dep.group('name'),
                    version=parsed_dep.group('version')
                )
            )
            continue

    gomods.require = require
    gomods.exclude = exclude

    return gomods


#######################################################################################
"""
This part handles go.sum files from Go.
See https://blog.golang.org/using-go-modules for details

A go.sum file contains pinned Go modules checksums of two styles:

For example:
github.com/BurntSushi/toml v0.3.1 h1:WXkYYl6Yr3qBf1K79EBnL4mak0OimBfB0XUf9Vl28OQ=
github.com/BurntSushi/toml v0.3.1/go.mod h1:xHWCNGjB5oqiDr8zfno3MHue2Ht5sIBksp03qcyfWMU=

... where the line with /go.mod is for a check of that go.mod file 
and the other line contains a dirhash for that path as documented as
https://pkg.go.dev/golang.org/x/mod/sumdb/dirhash

Here are some example of usage of this module::

>>> p = get_dependency('github.com/BurntSushi/toml v0.3.1 h1:WXkYYl6Yr3qBf1K79EBnL4mak0OimBfB0XUf9Vl28OQ=')
>>> assert p.group('namespace') == ('github.com/BurntSushi')
>>> assert p.group('name') == ('toml')
>>> assert p.group('version') == ('v0.3.1')
>>> assert p.group('checksum') == ('WXkYYl6Yr3qBf1K79EBnL4mak0OimBfB0XUf9Vl28OQ=')
"""

# Regex expressions to parse go.sum file dependency
# dep example: github.com/BurntSushi/toml v0.3.1 h1:WXkYY....
get_dependency = re.compile(
    r'(?P<namespace>[^\s]*)'
    r'\/'
    r'(?P<name>[^\s]*)'
    r'\s*'
    r'(?P<version>[^\s]*)'
    r'\s*'
    r'h1:(?P<checksum>[^\s]*)\s*'
).match


def parse_gosum(location):
    """
    Return a list of GoSum from parsing the go.sum file at `location`.
    """
    with io.open(location, encoding='utf-8', closefd=True) as data:
        lines = data.readlines()

    gosums = []

    for line in lines:
        line = line.replace('/go.mod', '')
        parsed_dep = get_dependency(line)

        dep = GoModule(
                namespace=parsed_dep.group('namespace'),
                name=parsed_dep.group('name'),
                version=parsed_dep.group('version')
            )

        if dep in gosums:
            continue

        gosums.append(dep)

    return gosums
